# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import unittest
import aniso8601

from aniso8601.builders import (DateTuple, DatetimeTuple,
                                DurationTuple, IntervalTuple, TimeTuple,
                                TimezoneTuple)
from aniso8601.exceptions import ISOFormatError
from aniso8601.interval import (_get_interval_component_resolution,
                                _get_interval_resolution,
                                _parse_interval, _parse_interval_end,
                                get_interval_resolution,
                                get_repeating_interval_resolution,
                                parse_interval, parse_repeating_interval)
from aniso8601.resolution import IntervalResolution
from aniso8601.tests.compat import mock

class TestIntervalParser_UtilityFunctions(unittest.TestCase):
    def test_get_interval_resolution(self):
        self.assertEqual(_get_interval_resolution(IntervalTuple(start=DateTuple(YYYY='2001', MM='02', DD='03',
                                                                                Www=None, D=None, DDD=None),
                                                                end=DatetimeTuple(DateTuple(YYYY='2001', MM='02', DD='03',
                                                                                            Www=None, D=None, DDD=None),
                                                                                  TimeTuple(hh='04', mm='05', ss='06',
                                                                                            tz=None)),
                                                                duration=None)),
                                                  IntervalResolution.Seconds)
        self.assertEqual(_get_interval_resolution(IntervalTuple(start=DatetimeTuple(DateTuple(YYYY='2001', MM='02', DD='03',
                                                                                              Www=None, D=None, DDD=None),
                                                                                    TimeTuple(hh='04', mm='05', ss='06',
                                                                                              tz=None)),
                                                                end=DateTuple(YYYY='2001', MM='02', DD='03',
                                                                              Www=None, D=None, DDD=None),
                                                                duration=None)),
                                                  IntervalResolution.Seconds)

        self.assertEqual(_get_interval_resolution(IntervalTuple(start=DateTuple(YYYY='2001', MM='02', DD='03',
                                                                                Www=None, D=None, DDD=None),
                                                                end=None,
                                                                duration=DurationTuple(PnY='1', PnM='2', PnW=None, PnD='3',
                                                                                       TnH='4', TnM='5', TnS='6'))),
                                                  IntervalResolution.Seconds)
        self.assertEqual(_get_interval_resolution(IntervalTuple(start=DatetimeTuple(DateTuple(YYYY='2001', MM='02', DD='03',
                                                                                              Www=None, D=None, DDD=None),
                                                                                    TimeTuple(hh='04', mm='05', ss='06',
                                                                                              tz=None)),
                                                                end=None,
                                                                duration=DurationTuple(PnY='1', PnM='2', PnW=None, PnD='3',
                                                                                       TnH=None, TnM=None, TnS=None))),
                                                  IntervalResolution.Seconds)

        self.assertEqual(_get_interval_resolution(IntervalTuple(start=None,
                                                                end=DateTuple(YYYY='2001', MM='02', DD='03',
                                                                              Www=None, D=None, DDD=None),
                                                                duration=DurationTuple(PnY='1', PnM='2', PnW=None, PnD='3',
                                                                                       TnH='4', TnM='5', TnS='6'))),
                                                  IntervalResolution.Seconds)
        self.assertEqual(_get_interval_resolution(IntervalTuple(start=None,
                                                                end=DatetimeTuple(DateTuple(YYYY='2001', MM='02', DD='03',
                                                                                            Www=None, D=None, DDD=None),
                                                                                  TimeTuple(hh='04', mm='05', ss='06',
                                                                                            tz=None)),
                                                                duration=DurationTuple(PnY='1', PnM='2', PnW=None, PnD='3',
                                                                                       TnH=None, TnM=None, TnS=None))),
                                                  IntervalResolution.Seconds)

    def test_get_interval_component_resolution(self):
        self.assertEqual(_get_interval_component_resolution(DateTuple(YYYY='2001', MM=None, DD=None,
                                                                      Www=None, D=None, DDD='123')),
                         IntervalResolution.Ordinal)
        self.assertEqual(_get_interval_component_resolution(DateTuple(YYYY='2001', MM=None, DD=None,
                                                                      Www='12', D='3', DDD=None)),
                         IntervalResolution.Weekday)
        self.assertEqual(_get_interval_component_resolution(DateTuple(YYYY='2001', MM=None, DD=None,
                                                                      Www='12', D=None, DDD=None)),
                         IntervalResolution.Week)
        self.assertEqual(_get_interval_component_resolution(DateTuple(YYYY='2001', MM='02', DD='03',
                                                                      Www=None, D=None, DDD=None)),
                         IntervalResolution.Day)
        self.assertEqual(_get_interval_component_resolution(DateTuple(YYYY='2001', MM='02', DD=None,
                                                                      Www=None, D=None, DDD=None)),
                         IntervalResolution.Month)
        self.assertEqual(_get_interval_component_resolution(DateTuple(YYYY='2001', MM=None, DD=None,
                                                                      Www=None, D=None, DDD=None)),
                         IntervalResolution.Year)

        self.assertEqual(_get_interval_component_resolution(DatetimeTuple(DateTuple(YYYY='2001', MM='02', DD='03',
                                                                                    Www=None, D=None, DDD=None),
                                                                          TimeTuple(hh='04', mm='05', ss='06',
                                                                                    tz=None))),
                         IntervalResolution.Seconds)
        self.assertEqual(_get_interval_component_resolution(DatetimeTuple(DateTuple(YYYY='2001', MM='02', DD='03',
                                                                                    Www=None, D=None, DDD=None),
                                                                          TimeTuple(hh='04', mm='05', ss=None,
                                                                                    tz=None))),
                         IntervalResolution.Minutes)
        self.assertEqual(_get_interval_component_resolution(DatetimeTuple(DateTuple(YYYY='2001', MM='02', DD='03',
                                                                                    Www=None, D=None, DDD=None),
                                                                          TimeTuple(hh='04', mm=None, ss=None,
                                                                                    tz=None))),
                         IntervalResolution.Hours)

        self.assertEqual(_get_interval_component_resolution(DurationTuple(PnY='1', PnM='2', PnW=None, PnD='3',
                                                                          TnH='4', TnM='5', TnS='6')),
                         IntervalResolution.Seconds)
        self.assertEqual(_get_interval_component_resolution(DurationTuple(PnY='1', PnM='2', PnW=None, PnD='3',
                                                                          TnH='4', TnM='5', TnS=None)),
                         IntervalResolution.Minutes)
        self.assertEqual(_get_interval_component_resolution(DurationTuple(PnY='1', PnM='2', PnW=None, PnD='3',
                                                                          TnH='4', TnM=None, TnS=None)),
                         IntervalResolution.Hours)
        self.assertEqual(_get_interval_component_resolution(DurationTuple(PnY='1', PnM='2', PnW=None, PnD='3',
                                                                          TnH=None, TnM=None, TnS=None)),
                         IntervalResolution.Day)
        self.assertEqual(_get_interval_component_resolution(DurationTuple(PnY='1', PnM='2', PnW=None, PnD=None,
                                                                          TnH=None, TnM=None, TnS=None)),
                         IntervalResolution.Month)
        self.assertEqual(_get_interval_component_resolution(DurationTuple(PnY='1', PnM=None, PnW=None, PnD=None,
                                                                          TnH=None, TnM=None, TnS=None)),
                         IntervalResolution.Year)
        self.assertEqual(_get_interval_component_resolution(DurationTuple(PnY=None, PnM=None, PnW='3', PnD=None,
                                                                          TnH=None, TnM=None, TnS=None)),
                         IntervalResolution.Week)

class TestIntervalParserFunctions(unittest.TestCase):
    def test_get_interval_resolution_date(self):
        self.assertEqual(get_interval_resolution('P1.5Y/2018'),
                         IntervalResolution.Year)
        self.assertEqual(get_interval_resolution('P1.5Y/2018-03'),
                         IntervalResolution.Month)
        self.assertEqual(get_interval_resolution('P1.5Y/2018-03-06'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('P1.5Y/2018W01'),
                         IntervalResolution.Week)
        self.assertEqual(get_interval_resolution('P1.5Y/2018-306'),
                         IntervalResolution.Ordinal)
        self.assertEqual(get_interval_resolution('P1.5Y/2018W012'),
                         IntervalResolution.Weekday)

        self.assertEqual(get_interval_resolution('2018/P1.5Y'),
                         IntervalResolution.Year)
        self.assertEqual(get_interval_resolution('2018-03/P1.5Y'),
                         IntervalResolution.Month)
        self.assertEqual(get_interval_resolution('2018-03-06/P1.5Y'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('2018W01/P1.5Y'),
                         IntervalResolution.Week)
        self.assertEqual(get_interval_resolution('2018-306/P1.5Y'),
                         IntervalResolution.Ordinal)
        self.assertEqual(get_interval_resolution('2018W012/P1.5Y'),
                         IntervalResolution.Weekday)

    def test_get_interval_resolution_time(self):
        self.assertEqual(get_interval_resolution('P1M/1981-04-05T01'),
                         IntervalResolution.Hours)
        self.assertEqual(get_interval_resolution('P1M/1981-04-05T01:01'),
                         IntervalResolution.Minutes)
        self.assertEqual(get_interval_resolution('P1M/1981-04-05T01:01:00'),
                         IntervalResolution.Seconds)

        self.assertEqual(get_interval_resolution('1981-04-05T01/P1M'),
                         IntervalResolution.Hours)
        self.assertEqual(get_interval_resolution('1981-04-05T01:01/P1M'),
                         IntervalResolution.Minutes)
        self.assertEqual(get_interval_resolution('1981-04-05T01:01:00/P1M'),
                         IntervalResolution.Seconds)

    def test_get_interval_resolution_duration(self):
        self.assertEqual(get_interval_resolution('2014-11-12/P1Y2M3D'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('2014-11-12/P1Y2M'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('2014-11-12/P1Y'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('2014-11-12/P1W'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('2014-11-12/P1Y2M3DT4H'),
                         IntervalResolution.Hours)
        self.assertEqual(get_interval_resolution('2014-11-12/P1Y2M3DT4H54M'),
                         IntervalResolution.Minutes)
        self.assertEqual(get_interval_resolution('2014-11-12/P1Y2M3DT4H54M6S'),
                         IntervalResolution.Seconds)

        self.assertEqual(get_interval_resolution('P1Y2M3D/2014-11-12'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('P1Y2M/2014-11-12'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('P1Y/2014-11-12'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('P1W/2014-11-12'),
                         IntervalResolution.Day)
        self.assertEqual(get_interval_resolution('P1Y2M3DT4H/2014-11-12'),
                         IntervalResolution.Hours)
        self.assertEqual(get_interval_resolution('P1Y2M3DT4H54M/2014-11-12'),
                         IntervalResolution.Minutes)
        self.assertEqual(get_interval_resolution('P1Y2M3DT4H54M6S/2014-11-12'),
                         IntervalResolution.Seconds)

    def test_parse_interval(self):
        testtuples = (('P1M/1981-04-05T01:01:00',
                       {'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None)),
                        'duration': DurationTuple(None, '1', None, None,
                                                  None, None, None)}),
                      ('P1M/1981-04-05',
                       {'end': DateTuple('1981', '04', '05', None, None, None),
                        'duration': DurationTuple(None, '1', None, None,
                                                  None, None, None)}),
                      ('P1,5Y/2018-03-06',
                       {'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple('1.5', None, None, None,
                                                  None, None, None)}),
                      ('P1.5Y/2018-03-06',
                       {'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple('1.5', None, None, None,
                                                  None, None, None)}),
                      ('PT1H/2014-11-12',
                       {'end': DateTuple('2014', '11', '12', None, None, None),
                        'duration': DurationTuple(None, None, None, None,
                                                  '1', None, None)}),
                      ('PT4H54M6.5S/2014-11-12',
                       {'end': DateTuple('2014', '11', '12', None, None, None),
                        'duration': DurationTuple(None, None, None, None,
                                                  '4', '54', '6.5')}),
                      ('PT10H/2050-03-01T13:00:00Z',
                       {'end': DatetimeTuple(DateTuple('2050', '03', '01',
                                                       None, None, None),
                                             TimeTuple('13', '00', '00',
                                                       TimezoneTuple(False,
                                                                     True,
                                                                     None,
                                                                     None,
                                                                     'Z'))),
                        'duration': DurationTuple(None, None, None,
                                                  None, '10', None, None)}),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      ('PT0.0000001S/2018-03-06',
                       {'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '0.0000001')}),
                      ('PT2.0000048S/2018-03-06',
                       {'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '2.0000048')}),
                      ('1981-04-05T01:01:00/P1M1DT1M',
                       {'start': DatetimeTuple(DateTuple('1981', '04', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01',
                                                         '00', None)),
                        'duration': DurationTuple(None, '1', None,
                                                  '1', None, '1', None)}),
                      ('1981-04-05/P1M1D',
                       {'start': DateTuple('1981', '04', '05',
                                           None, None, None),
                        'duration': DurationTuple(None, '1', None,
                                                  '1', None, None, None)}),
                      ('2018-03-06/P2,5M',
                       {'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, '2.5', None,
                                                  None, None, None, None)}),
                      ('2018-03-06/P2.5M',
                       {'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, '2.5', None,
                                                  None, None, None, None)}),
                      ('2014-11-12/PT1H',
                       {'start': DateTuple('2014', '11', '12',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, '1', None, None)}),
                      ('2014-11-12/PT4H54M6.5S',
                       {'start': DateTuple('2014', '11', '12',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, '4', '54', '6.5')}),
                      ('2050-03-01T13:00:00Z/PT10H',
                       {'start': DatetimeTuple(DateTuple('2050', '03', '01',
                                                         None, None, None),
                                               TimeTuple('13', '00', '00',
                                                         TimezoneTuple(False,
                                                                       True,
                                                                       None,
                                                                       None,
                                                                       'Z'))),
                        'duration': DurationTuple(None, None, None,
                                                  None, '10', None, None)}),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      ('2018-03-06/PT0.0000001S',
                       {'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '0.0000001')}),
                      ('2018-03-06/PT2.0000048S',
                       {'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '2.0000048')}),
                      ('1980-03-05T01:01:00/1981-04-05T01:01:00',
                       {'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00',
                                                         None)),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None))}),
                      ('1980-03-05T01:01:00/1981-04-05',
                       {'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                  TimeTuple('01', '01', '00',
                                            None)),
                        'end': DateTuple('1981', '04', '05',
                                         None, None, None)}),
                      ('1980-03-05/1981-04-05T01:01:00',
                       {'start': DateTuple('1980', '03', '05',
                                           None, None, None),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None))}),
                      ('1980-03-05/1981-04-05',
                       {'start': DateTuple('1980', '03', '05',
                                           None, None, None),
                        'end': DateTuple('1981', '04', '05',
                                         None, None, None)}),
                      ('1981-04-05/1980-03-05',
                       {'start': DateTuple('1981', '04', '05',
                                           None, None, None),
                        'end': DateTuple('1980', '03', '05',
                                         None, None, None)}),
                      ('2050-03-01T13:00:00Z/2050-05-11T15:30:00Z',
                       {'start': DatetimeTuple(DateTuple('2050', '03', '01',
                                                         None, None, None),
                                               TimeTuple('13', '00', '00',
                                               TimezoneTuple(False, True,
                                                             None, None,
                                                             'Z'))),
                        'end': DatetimeTuple(DateTuple('2050', '05', '11',
                                                       None, None, None),
                                             TimeTuple('15', '30', '00',
                                             TimezoneTuple(False, True,
                                                           None, None,
                                                           'Z')))}),
                      #Test concise interval
                      ('2020-01-01/02',
                       {'start': DateTuple('2020', '01', '01',
                                           None, None, None),
                        'end': DateTuple(None, None, '02',
                                         None, None, None)}),
                      ('2008-02-15/03-14',
                       {'start': DateTuple('2008', '02', '15',
                                           None, None, None),
                        'end': DateTuple(None, '03', '14',
                                         None, None, None)}),
                      ('2007-12-14T13:30/15:30',
                       {'start': DatetimeTuple(DateTuple('2007', '12', '14',
                                                         None, None, None),
                                               TimeTuple('13', '30', None,
                                                         None)),
                        'end': TimeTuple('15', '30', None,
                                         None)}),
                      ('2007-11-13T09:00/15T17:00',
                       {'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('09', '00', None,
                                                         None)),
                        'end': DatetimeTuple(DateTuple(None, None, '15',
                                                       None, None, None),
                                             TimeTuple('17', '00', None,
                                                       None))}),
                      ('2007-11-13T00:00/16T00:00',
                       {'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('00', '00', None,
                                                         None)),
                        'end': DatetimeTuple(DateTuple(None, None, '16',
                                                       None, None, None),
                                             TimeTuple('00', '00', None,
                                                       None))}),
                      ('2007-11-13T09:00Z/15T17:00',
                       {'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('09', '00', None,
                                                         TimezoneTuple(False,
                                                                       True,
                                                                       None,
                                                                       None,
                                                                       'Z'))),
                        'end': DatetimeTuple(DateTuple(None, None, '15',
                                                       None, None, None),
                                             TimeTuple('17', '00', None,
                                                       None))}),
                      ('2007-11-13T00:00/12:34.567',
                       {'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('00', '00', None,
                                                         None)),
                        'end': TimeTuple('12', '34.567', None,
                                         None)}),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      ('1980-03-05T01:01:00.0000001/'
                       '1981-04-05T14:43:59.9999997',
                       {'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01',
                                                         '00.0000001',
                                                         None)),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('14', '43',
                                                       '59.9999997', None))}))

        for testtuple in testtuples:
            with mock.patch.object(aniso8601.interval.PythonTimeBuilder,
                                   'build_interval') as mockBuildInterval:
                mockBuildInterval.return_value = testtuple[1]

                result = parse_interval(testtuple[0])

                self.assertEqual(result, testtuple[1])
                mockBuildInterval.assert_called_once_with(**testtuple[1])

        #Test different separators
        with mock.patch.object(aniso8601.interval.PythonTimeBuilder,
                               'build_interval') as mockBuildInterval:
            expectedargs = {'start': DatetimeTuple(DateTuple('1980', '03',
                                                             '05', None,
                                                             None, None),
                                                   TimeTuple('01', '01',
                                                             '00', None)),
                            'end': DatetimeTuple(DateTuple('1981', '04',
                                                           '05', None,
                                                           None, None),
                                                 TimeTuple('01', '01',
                                                           '00', None))}

            mockBuildInterval.return_value = expectedargs

            result = parse_interval('1980-03-05T01:01:00--1981-04-05T01:01:00',
                                    intervaldelimiter='--')

            self.assertEqual(result, expectedargs)
            mockBuildInterval.assert_called_once_with(**expectedargs)

        with mock.patch.object(aniso8601.interval.PythonTimeBuilder,
                               'build_interval') as mockBuildInterval:
            expectedargs = {'start': DatetimeTuple(DateTuple('1980', '03',
                                                             '05', None,
                                                             None, None),
                                      TimeTuple('01', '01', '00', None)),
                            'end': DatetimeTuple(DateTuple('1981', '04',
                                                           '05', None,
                                                           None, None),
                                                 TimeTuple('01', '01',
                                                           '00', None))}

            mockBuildInterval.return_value = expectedargs

            result = parse_interval('1980-03-05 01:01:00/1981-04-05 01:01:00',
                                    datetimedelimiter=' ')

            self.assertEqual(result, expectedargs)
            mockBuildInterval.assert_called_once_with(**expectedargs)

    def test_parse_interval_mockbuilder(self):
        mockBuilder = mock.Mock()

        expectedargs = {'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None)),
                        'duration': DurationTuple(None, '1', None, None,
                                                  None, None, None)}

        mockBuilder.build_interval.return_value = expectedargs

        result = parse_interval('P1M/1981-04-05T01:01:00', builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

        mockBuilder = mock.Mock()

        expectedargs = {'start': DateTuple('2014', '11', '12',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None, None,
                                                  '1', None, None)}

        mockBuilder.build_interval.return_value = expectedargs

        result = parse_interval('2014-11-12/PT1H', builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

        mockBuilder = mock.Mock()

        expectedargs = {'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00',
                                                         None)),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None))}

        mockBuilder.build_interval.return_value = expectedargs

        result = parse_interval('1980-03-05T01:01:00/1981-04-05T01:01:00',
                       builder=mockBuilder)

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

    def test_parse_interval_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                parse_interval(testtuple, builder=None)

    def test_parse_interval_baddelimiter(self):
        testtuples = ('1980-03-05T01:01:00,1981-04-05T01:01:00',
                      'P1M 1981-04-05T01:01:00')

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_interval(testtuple, builder=None)

    def test_parse_interval_badstr(self):
        testtuples = ('/', '0/0/0', '20.50230/0', '5/%', '1/21', 'bad', '')

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_interval(testtuple, builder=None)

    def test_parse_interval_repeating(self):
        #Parse interval can't parse repeating intervals
        with self.assertRaises(ISOFormatError):
            parse_interval('R3/1981-04-05/P1D')

        with self.assertRaises(ISOFormatError):
            parse_interval('R3/1981-04-05/P0003-06-04T12:30:05.5')

        with self.assertRaises(ISOFormatError):
            parse_interval('R/PT1H2M/1980-03-05T01:01:00')

    def test_parse_interval_suffixgarbage(self):
        #Don't allow garbage after the duration
        #https://bitbucket.org/nielsenb/aniso8601/issues/9/durations-with-trailing-garbage-are-parsed
        with self.assertRaises(ISOFormatError):
            parse_interval('2001/P1Dasdf', builder=None)

        with self.assertRaises(ISOFormatError):
            parse_interval('P1Dasdf/2001', builder=None)

        with self.assertRaises(ISOFormatError):
            parse_interval('2001/P0003-06-04T12:30:05.5asdfasdf', builder=None)

        with self.assertRaises(ISOFormatError):
            parse_interval('P0003-06-04T12:30:05.5asdfasdf/2001', builder=None)

    def test_parse_interval_internal(self):
        #Test the internal _parse_interval function
        testtuples = (('P1M/1981-04-05T01:01:00',
                       {'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00', None)),
                        'duration': DurationTuple(None, '1', None, None,
                                                  None, None, None)}),
                      ('P1M/1981-04-05',
                       {'end': DateTuple('1981', '04', '05', None, None, None),
                        'duration': DurationTuple(None, '1', None, None,
                                                  None, None, None)}),
                      ('P1,5Y/2018-03-06',
                       {'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple('1.5', None, None, None,
                                                  None, None, None)}),
                      ('P1.5Y/2018-03-06',
                       {'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple('1.5', None, None, None,
                                                  None, None, None)}),
                      ('PT1H/2014-11-12',
                       {'end': DateTuple('2014', '11', '12', None, None, None),
                        'duration': DurationTuple(None, None, None, None,
                                                  '1', None, None)}),
                      ('PT4H54M6.5S/2014-11-12',
                       {'end': DateTuple('2014', '11', '12', None, None, None),
                        'duration': DurationTuple(None, None, None, None,
                                                  '4', '54', '6.5')}),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      ('PT0.0000001S/2018-03-06',
                       {'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '0.0000001')}),
                      ('PT2.0000048S/2018-03-06',
                       {'end': DateTuple('2018', '03', '06', None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '2.0000048')}),
                      ('1981-04-05T01:01:00/P1M1DT1M',
                       {'start': DatetimeTuple(DateTuple('1981', '04', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00', None)),
                        'duration': DurationTuple(None, '1', None,
                                                  '1', None, '1', None)}),
                      ('1981-04-05/P1M1D',
                       {'start': DateTuple('1981', '04', '05',
                                           None, None, None),
                        'duration': DurationTuple(None, '1', None,
                                                  '1', None, None, None)}),
                      ('2018-03-06/P2,5M',
                       {'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, '2.5', None,
                                                  None, None, None, None)}),
                      ('2018-03-06/P2.5M',
                       {'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, '2.5', None,
                                                  None, None, None, None)}),
                      ('2014-11-12/PT1H',
                       {'start': DateTuple('2014', '11', '12',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, '1', None, None)}),
                      ('2014-11-12/PT4H54M6.5S',
                       {'start': DateTuple('2014', '11', '12',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, '4', '54', '6.5')}),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      ('2018-03-06/PT0.0000001S',
                       {'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '0.0000001')}),
                      ('2018-03-06/PT2.0000048S',
                       {'start': DateTuple('2018', '03', '06',
                                           None, None, None),
                        'duration': DurationTuple(None, None, None,
                                                  None, None, None,
                                                  '2.0000048')}),
                      ('1980-03-05T01:01:00/1981-04-05T01:01:00',
                       {'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00',
                                                         None)),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None))}),
                      ('1980-03-05T01:01:00/1981-04-05',
                       {'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00',
                                                         None)),
                        'end': DateTuple('1981', '04', '05',
                                         None, None, None)}),
                      ('1980-03-05/1981-04-05T01:01:00',
                       {'start': DateTuple('1980', '03', '05',
                                           None, None, None),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None))}),
                      ('1980-03-05/1981-04-05',
                       {'start': DateTuple('1980', '03', '05',
                                           None, None, None),
                        'end': DateTuple('1981', '04', '05',
                                         None, None, None)}),
                      ('1981-04-05/1980-03-05',
                       {'start': DateTuple('1981', '04', '05',
                                           None, None, None),
                        'end': DateTuple('1980', '03', '05',
                                         None, None, None)}),
                      #Test concise interval
                      ('2020-01-01/02',
                       {'start': DateTuple('2020', '01', '01',
                                           None, None, None),
                        'end': DateTuple(None, None, '02',
                                         None, None, None)}),
                      ('2008-02-15/03-14',
                       {'start': DateTuple('2008', '02', '15',
                                           None, None, None),
                        'end': DateTuple(None, '03', '14',
                                         None, None, None)}),
                      ('2007-12-14T13:30/15:30',
                       {'start': DatetimeTuple(DateTuple('2007', '12', '14',
                                                         None, None, None),
                                               TimeTuple('13', '30', None,
                                                         None)),
                        'end': TimeTuple('15', '30', None,
                                         None)}),
                      ('2007-11-13T09:00/15T17:00',
                       {'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('09', '00', None,
                                                         None)),
                        'end': DatetimeTuple(DateTuple(None, None, '15',
                                                       None, None, None),
                                             TimeTuple('17', '00', None,
                                                       None))}),
                      ('2007-11-13T00:00/16T00:00',
                       {'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('00', '00', None,
                                                         None)),
                        'end': DatetimeTuple(DateTuple(None, None, '16',
                                                       None, None, None),
                                             TimeTuple('00', '00', None,
                                                       None))}),
                      ('2007-11-13T09:00Z/15T17:00',
                       {'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('09', '00', None,
                                                         TimezoneTuple(False,
                                                                       True,
                                                                       None,
                                                                       None,
                                                                       'Z'))),
                        'end': DatetimeTuple(DateTuple(None, None, '15',
                                                       None, None, None),
                                             TimeTuple('17', '00', None,
                                                       None))}),
                      ('2007-11-13T00:00/12:34.567',
                       {'start': DatetimeTuple(DateTuple('2007', '11', '13',
                                                         None, None, None),
                                               TimeTuple('00', '00', None,
                                                         None)),
                        'end': TimeTuple('12', '34.567', None,
                                         None)}),
                      #Make sure we truncate, not round
                      #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                      ('1980-03-05T01:01:00.0000001/'
                       '1981-04-05T14:43:59.9999997',
                       {'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00.0000001',
                                                         None)),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('14', '43', '59.9999997',
                                                       None))}))

        for testtuple in testtuples:
            mockBuilder = mock.Mock()
            mockBuilder.build_interval.return_value = testtuple[1]

            result = _parse_interval(testtuple[0], mockBuilder)

            self.assertEqual(result, testtuple[1])
            mockBuilder.build_interval.assert_called_once_with(**testtuple[1])

        #Test different separators
        expectedargs = {'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                               TimeTuple('01', '01', '00',
                                                         None)),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None))}

        mockBuilder = mock.Mock()
        mockBuilder.build_interval.return_value = expectedargs

        result = _parse_interval('1980-03-05T01:01:00--1981-04-05T01:01:00',
                                 mockBuilder,
                                 intervaldelimiter='--')

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

        expectedargs = {'start': DatetimeTuple(DateTuple('1980', '03', '05',
                                                         None, None, None),
                                  TimeTuple('01', '01', '00',
                                            None)),
                        'end': DatetimeTuple(DateTuple('1981', '04', '05',
                                                       None, None, None),
                                             TimeTuple('01', '01', '00',
                                                       None))}

        mockBuilder = mock.Mock()
        mockBuilder.build_interval.return_value = expectedargs

        _parse_interval('1980-03-05 01:01:00/1981-04-05 01:01:00',
                        mockBuilder,
                        datetimedelimiter=' ')

        self.assertEqual(result, expectedargs)
        mockBuilder.build_interval.assert_called_once_with(**expectedargs)

    def test_parse_interval_end(self):
        self.assertEqual(_parse_interval_end('02', DateTuple('2020', '01', '01',
                                                             None, None, None),
                                             'T'),
                                             DateTuple(None, None, '02',
                                                       None, None, None))

        self.assertEqual(_parse_interval_end('03-14', DateTuple('2008', '02', '15',
                                                                None, None, None),
                                             'T'),
                                             DateTuple(None, '03', '14',
                                                       None, None, None))

        self.assertEqual(_parse_interval_end('0314', DateTuple('2008', '02', '15',
                                                                None, None, None),
                                             'T'),
                                             DateTuple(None, '03', '14',
                                                       None, None, None))

        self.assertEqual(_parse_interval_end('15:30', DatetimeTuple(DateTuple('2007', '12', '14',
                                                                              None, None, None),
                                                                    TimeTuple('13', '30', None,
                                                                              None)),
                                             'T'),
                                             TimeTuple('15', '30', None,
                                                       None))

        self.assertEqual(_parse_interval_end('15T17:00', DatetimeTuple(DateTuple('2007', '11', '13',
                                                                                 None, None, None),
                                                                       TimeTuple('09', '00', None,
                                                                                 None)),
                                             'T'),
                                             DatetimeTuple(DateTuple(None, None, '15',
                                                                     None, None, None),
                                                           TimeTuple('17', '00', None,
                                                                     None)))

        self.assertEqual(_parse_interval_end('16T00:00', DatetimeTuple(DateTuple('2007', '11', '13',
                                                                                 None, None, None),
                                                                       TimeTuple('00', '00', None,
                                                                                 None)),
                                             'T'),
                                             DatetimeTuple(DateTuple(None, None, '16',
                                                                     None, None, None),
                                                           TimeTuple('00', '00', None,
                                                                     None)))

        self.assertEqual(_parse_interval_end('15 17:00', DatetimeTuple(DateTuple('2007', '11', '13',
                                                                                 None, None, None),
                                                                       TimeTuple('09', '00', None,
                                                                                 None)),
                                             ' '),
                                             DatetimeTuple(DateTuple(None, None, '15',
                                                                     None, None, None),
                                                           TimeTuple('17', '00', None,
                                                                     None)))

        self.assertEqual(_parse_interval_end('12:34.567', DatetimeTuple(DateTuple('2007', '11', '13',
                                                                                 None, None, None),
                                                                       TimeTuple('00', '00', None,
                                                                                 None)),
                                             'T'),
                                             TimeTuple('12', '34.567', None,
                                                       None))

class TestRepeatingIntervalParserFunctions(unittest.TestCase):
    def test_get_interval_resolution_date(self):
        self.assertEqual(get_repeating_interval_resolution('R/P1.5Y/2018'),
                         IntervalResolution.Year)
        self.assertEqual(get_repeating_interval_resolution('R1/P1.5Y/2018-03'),
                         IntervalResolution.Month)
        self.assertEqual(get_repeating_interval_resolution('R2/P1.5Y/2018-03-06'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R3/P1.5Y/2018W01'),
                         IntervalResolution.Week)
        self.assertEqual(get_repeating_interval_resolution('R4/P1.5Y/2018-306'),
                         IntervalResolution.Ordinal)
        self.assertEqual(get_repeating_interval_resolution('R5/P1.5Y/2018W012'),
                         IntervalResolution.Weekday)

        self.assertEqual(get_repeating_interval_resolution('R/2018/P1.5Y'),
                         IntervalResolution.Year)
        self.assertEqual(get_repeating_interval_resolution('R1/2018-03/P1.5Y'),
                         IntervalResolution.Month)
        self.assertEqual(get_repeating_interval_resolution('R2/2018-03-06/P1.5Y'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R3/2018W01/P1.5Y'),
                         IntervalResolution.Week)
        self.assertEqual(get_repeating_interval_resolution('R4/2018-306/P1.5Y'),
                         IntervalResolution.Ordinal)
        self.assertEqual(get_repeating_interval_resolution('R5/2018W012/P1.5Y'),
                         IntervalResolution.Weekday)

    def test_get_interval_resolution_time(self):
        self.assertEqual(get_repeating_interval_resolution('R/P1M/1981-04-05T01'),
                         IntervalResolution.Hours)
        self.assertEqual(get_repeating_interval_resolution('R1/P1M/1981-04-05T01:01'),
                         IntervalResolution.Minutes)
        self.assertEqual(get_repeating_interval_resolution('R2/P1M/1981-04-05T01:01:00'),
                         IntervalResolution.Seconds)

        self.assertEqual(get_repeating_interval_resolution('R/1981-04-05T01/P1M'),
                         IntervalResolution.Hours)
        self.assertEqual(get_repeating_interval_resolution('R1/1981-04-05T01:01/P1M'),
                         IntervalResolution.Minutes)
        self.assertEqual(get_repeating_interval_resolution('R2/1981-04-05T01:01:00/P1M'),
                         IntervalResolution.Seconds)

    def test_get_interval_resolution_duration(self):
        self.assertEqual(get_repeating_interval_resolution('R/2014-11-12/P1Y2M3D'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R1/2014-11-12/P1Y2M'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R2/2014-11-12/P1Y'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R3/2014-11-12/P1W'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R4/2014-11-12/P1Y2M3DT4H'),
                         IntervalResolution.Hours)
        self.assertEqual(get_repeating_interval_resolution('R5/2014-11-12/P1Y2M3DT4H54M'),
                         IntervalResolution.Minutes)
        self.assertEqual(get_repeating_interval_resolution('R6/2014-11-12/P1Y2M3DT4H54M6S'),
                         IntervalResolution.Seconds)

        self.assertEqual(get_repeating_interval_resolution('R/P1Y2M3D/2014-11-12'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R1/P1Y2M/2014-11-12'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R2/P1Y/2014-11-12'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R3/P1W/2014-11-12'),
                         IntervalResolution.Day)
        self.assertEqual(get_repeating_interval_resolution('R4/P1Y2M3DT4H/2014-11-12'),
                         IntervalResolution.Hours)
        self.assertEqual(get_repeating_interval_resolution('R5/P1Y2M3DT4H54M/2014-11-12'),
                         IntervalResolution.Minutes)
        self.assertEqual(get_repeating_interval_resolution('R6/P1Y2M3DT4H54M6S/2014-11-12'),
                         IntervalResolution.Seconds)

    def test_parse_repeating_interval(self):
        with mock.patch.object(aniso8601.interval.PythonTimeBuilder,
                               'build_repeating_interval') as mockBuilder:
            expectedargs = {'R': False, 'Rnn': '3',
                            'interval': IntervalTuple(DateTuple('1981', '04',
                                                                '05', None,
                                                                None, None),
                                                      None,
                                                      DurationTuple(None, None,
                                                                    None, '1',
                                                                    None, None,
                                                                    None))}

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval('R3/1981-04-05/P1D')

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

        with mock.patch.object(aniso8601.interval.PythonTimeBuilder,
                               'build_repeating_interval') as mockBuilder:
            expectedargs = {'R': False, 'Rnn': '11',
                            'interval': IntervalTuple(None,
                                                      DatetimeTuple(DateTuple('1980', '03', '05',
                                                                              None, None, None),
                                                      TimeTuple('01', '01', '00',
                                                                None)),
                                                      DurationTuple(None, None, None,
                                                                    None, '1', '2',
                                                                    None))}

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval('R11/PT1H2M/1980-03-05T01:01:00')

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

        with mock.patch.object(aniso8601.interval.PythonTimeBuilder,
                               'build_repeating_interval') as mockBuilder:
            expectedargs = {'R': False, 'Rnn': '2',
                            'interval': IntervalTuple(DatetimeTuple(DateTuple('1980', '03', '05',
                                                                              None, None, None),
                                                      TimeTuple('01', '01', '00',
                                                                None)),
                                                      DatetimeTuple(DateTuple('1981', '04', '05',
                                                                              None, None, None),
                                                      TimeTuple('01', '01', '00',
                                                                None)),
                                                      None)}

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval('R2--1980-03-05T01:01:00--'
                                              '1981-04-05T01:01:00',
                                              intervaldelimiter='--')

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

        with mock.patch.object(aniso8601.interval.PythonTimeBuilder,
                               'build_repeating_interval') as mockBuilder:
            expectedargs = {'R': False, 'Rnn': '2',
                            'interval': IntervalTuple(DatetimeTuple(DateTuple('1980', '03', '05',
                                                                              None, None, None),
                                                                    TimeTuple('01', '01', '00',
                                                                              None)),
                                                      DatetimeTuple(DateTuple('1981', '04', '05',
                                                                              None, None, None),
                                                                    TimeTuple('01', '01', '00',
                                                                              None)),
                                                      None)}

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval('R2/'
                                              '1980-03-05 01:01:00/'
                                              '1981-04-05 01:01:00',
                                              datetimedelimiter=' ')

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

        with mock.patch.object(aniso8601.interval.PythonTimeBuilder,
                               'build_repeating_interval') as mockBuilder:
            expectedargs = {'R': True, 'Rnn': None,
                            'interval': IntervalTuple(None,
                                                      DatetimeTuple(DateTuple('1980', '03', '05',
                                                                              None, None, None),
                                                                    TimeTuple('01', '01', '00',
                                                                              None)),
                                                      DurationTuple(None, None, None,
                                                                    None, '1', '2',
                                                                    None))}

            mockBuilder.return_value = expectedargs

            result = parse_repeating_interval('R/PT1H2M/1980-03-05T01:01:00')

            self.assertEqual(result, expectedargs)
            mockBuilder.assert_called_once_with(**expectedargs)

    def test_parse_repeating_interval_mockbuilder(self):
        mockBuilder = mock.Mock()

        args = {'R': False, 'Rnn': '3',
                'interval': IntervalTuple(DateTuple('1981', '04', '05',
                                                    None, None, None),
                             None,
                             DurationTuple(None, None, None, '1',
                                           None, None, None))}

        mockBuilder.build_repeating_interval.return_value = args

        result = parse_repeating_interval('R3/1981-04-05/P1D',
                                          builder=mockBuilder)

        self.assertEqual(result, args)
        mockBuilder.build_repeating_interval.assert_called_once_with(**args)

        mockBuilder = mock.Mock()

        args = {'R': False, 'Rnn': '11',
                'interval': IntervalTuple(None,
                                          DatetimeTuple(DateTuple('1980',
                                                                  '03', '05',
                                                                  None, None,
                                                                  None),
                                                        TimeTuple('01', '01',
                                                                  '00', None)),
                                          DurationTuple(None, None, None, None,
                                                        '1', '2', None))}

        mockBuilder.build_repeating_interval.return_value = args

        result = parse_repeating_interval('R11/PT1H2M/1980-03-05T01:01:00',
                                          builder=mockBuilder)

        self.assertEqual(result, args)
        mockBuilder.build_repeating_interval.assert_called_once_with(**args)

        mockBuilder = mock.Mock()

        args = {'R': True, 'Rnn': None,
                'interval': IntervalTuple(None,
                                          DatetimeTuple(DateTuple('1980', '03',
                                                                  '05', None,
                                                                  None, None),
                                                        TimeTuple('01', '01',
                                                                  '00', None)),
                                                        DurationTuple(None,
                                                                      None,
                                                                      None,
                                                                      None,
                                                                      '1', '2',
                                                                      None))}

        mockBuilder.build_repeating_interval.return_value = args

        result = parse_repeating_interval('R/PT1H2M/1980-03-05T01:01:00',
                                          builder=mockBuilder)

        self.assertEqual(result, args)
        mockBuilder.build_repeating_interval.assert_called_once_with(**args)

    def test_parse_repeating_interval_badtype(self):
        testtuples = (None, 1, False, 1.234)

        for testtuple in testtuples:
            with self.assertRaises(ValueError):
                parse_repeating_interval(testtuple, builder=None)

    def test_parse_repeating_interval_baddelimiter(self):
        testtuples = ('R,PT1H2M,1980-03-05T01:01:00', 'R3 1981-04-05 P1D')

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_repeating_interval(testtuple, builder=None)

    def test_parse_repeating_interval_suffixgarbage(self):
        #Don't allow garbage after the duration
        #https://bitbucket.org/nielsenb/aniso8601/issues/9/durations-with-trailing-garbage-are-parsed
        with self.assertRaises(ISOFormatError):
            parse_repeating_interval('R3/1981-04-05/P1Dasdf', builder=None)

        with self.assertRaises(ISOFormatError):
            parse_repeating_interval('R3/'
                                     '1981-04-05/'
                                     'P0003-06-04T12:30:05.5asdfasdf',
                                     builder=None)

    def test_parse_repeating_interval_badstr(self):
        testtuples = ('bad', '')

        for testtuple in testtuples:
            with self.assertRaises(ISOFormatError):
                parse_repeating_interval(testtuple, builder=None)



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import datetime

from aniso8601.builders import (BaseTimeBuilder, DateTuple, DatetimeTuple,
                                Limit, TimeTuple, TupleBuilder, cast,
                                range_check)
from aniso8601.exceptions import (DayOutOfBoundsError,
                                  HoursOutOfBoundsError, ISOFormatError,
                                  LeapSecondError, MidnightBoundsError,
                                  MinutesOutOfBoundsError,
                                  MonthOutOfBoundsError,
                                  SecondsOutOfBoundsError,
                                  WeekOutOfBoundsError, YearOutOfBoundsError)
from aniso8601.utcoffset import UTCOffset
from collections import namedtuple
from functools import partial

DAYS_PER_YEAR = 365
DAYS_PER_MONTH = 30
DAYS_PER_WEEK = 7

HOURS_PER_DAY = 24

MINUTES_PER_HOUR = 60
MINUTES_PER_DAY = MINUTES_PER_HOUR * HOURS_PER_DAY

SECONDS_PER_MINUTE = 60
SECONDS_PER_DAY = MINUTES_PER_DAY * SECONDS_PER_MINUTE

MICROSECONDS_PER_SECOND = int(1e6)

MICROSECONDS_PER_MINUTE = 60 * MICROSECONDS_PER_SECOND
MICROSECONDS_PER_HOUR = 60 * MICROSECONDS_PER_MINUTE
MICROSECONDS_PER_DAY = 24 * MICROSECONDS_PER_HOUR
MICROSECONDS_PER_WEEK = 7 * MICROSECONDS_PER_DAY
MICROSECONDS_PER_MONTH = DAYS_PER_MONTH * MICROSECONDS_PER_DAY
MICROSECONDS_PER_YEAR = DAYS_PER_YEAR * MICROSECONDS_PER_DAY

TIMEDELTA_MAX_DAYS = datetime.timedelta.max.days

FractionalComponent = namedtuple('FractionalComponent', ['principal', 'microsecondremainder'])

def year_range_check(valuestr, limit):
    YYYYstr = valuestr

    #Truncated dates, like '19', refer to 1900-1999 inclusive,
    #we simply parse to 1900
    if len(valuestr) < 4:
        #Shift 0s in from the left to form complete year
        YYYYstr = valuestr.ljust(4, '0')

    return range_check(YYYYstr, limit)

def fractional_range_check(conversion, valuestr, limit):
    if valuestr is None:
        return None

    if '.' in valuestr:
        castfunc = partial(_cast_to_fractional_component, conversion)
    else:
        castfunc = int

    value = cast(valuestr, castfunc, thrownmessage=limit.casterrorstring)

    if type(value) is FractionalComponent:
        tocheck = float(valuestr)
    else:
        tocheck = int(valuestr)

    if limit.min is not None and tocheck < limit.min:
        raise limit.rangeexception(limit.rangeerrorstring)

    if limit.max is not None and tocheck > limit.max:
        raise limit.rangeexception(limit.rangeerrorstring)

    return value

def _cast_to_fractional_component(conversion, floatstr):
    #Splits a string with a decimal point into an int, and
    #int representing the floating point remainder as a number
    #of microseconds, determined by multiplying by conversion
    intpart, floatpart = floatstr.split('.')

    intvalue = int(intpart)
    preconvertedvalue = int(floatpart)

    convertedvalue = ((preconvertedvalue * conversion) //
                      (10 ** len(floatpart)))

    return FractionalComponent(intvalue, convertedvalue)

class PythonTimeBuilder(BaseTimeBuilder):
    #0000 (1 BC) is not representable as a Python date
    DATE_YYYY_LIMIT = Limit('Invalid year string.',
                            datetime.MINYEAR, datetime.MAXYEAR, YearOutOfBoundsError,
                            'Year must be between {0}..{1}.'
                            .format(datetime.MINYEAR, datetime.MAXYEAR),
                            year_range_check)
    TIME_HH_LIMIT = Limit('Invalid hour string.',
                          0, 24, HoursOutOfBoundsError,
                          'Hour must be between 0..24 with '
                          '24 representing midnight.',
                          partial(fractional_range_check, MICROSECONDS_PER_HOUR))
    TIME_MM_LIMIT = Limit('Invalid minute string.',
                          0, 59, MinutesOutOfBoundsError,
                          'Minute must be between 0..59.',
                          partial(fractional_range_check, MICROSECONDS_PER_MINUTE))
    TIME_SS_LIMIT = Limit('Invalid second string.',
                          0, 60, SecondsOutOfBoundsError,
                          'Second must be between 0..60 with '
                          '60 representing a leap second.',
                          partial(fractional_range_check, MICROSECONDS_PER_SECOND))
    DURATION_PNY_LIMIT = Limit('Invalid year duration string.',
                               None, None, YearOutOfBoundsError,
                               None,
                               partial(fractional_range_check, MICROSECONDS_PER_YEAR))
    DURATION_PNM_LIMIT = Limit('Invalid month duration string.',
                               None, None, MonthOutOfBoundsError,
                               None,
                               partial(fractional_range_check, MICROSECONDS_PER_MONTH))
    DURATION_PNW_LIMIT = Limit('Invalid week duration string.',
                               None, None, WeekOutOfBoundsError,
                               None,
                               partial(fractional_range_check, MICROSECONDS_PER_WEEK))
    DURATION_PND_LIMIT = Limit('Invalid day duration string.',
                               None, None, DayOutOfBoundsError,
                               None,
                               partial(fractional_range_check, MICROSECONDS_PER_DAY))
    DURATION_TNH_LIMIT = Limit('Invalid hour duration string.',
                               None, None, HoursOutOfBoundsError,
                               None,
                               partial(fractional_range_check, MICROSECONDS_PER_HOUR))
    DURATION_TNM_LIMIT = Limit('Invalid minute duration string.',
                               None, None, MinutesOutOfBoundsError,
                               None,
                               partial(fractional_range_check, MICROSECONDS_PER_MINUTE))
    DURATION_TNS_LIMIT = Limit('Invalid second duration string.',
                               None, None, SecondsOutOfBoundsError,
                               None,
                               partial(fractional_range_check, MICROSECONDS_PER_SECOND))

    DATE_RANGE_DICT = BaseTimeBuilder.DATE_RANGE_DICT
    DATE_RANGE_DICT['YYYY'] = DATE_YYYY_LIMIT

    TIME_RANGE_DICT = {'hh': TIME_HH_LIMIT, 'mm': TIME_MM_LIMIT, 'ss': TIME_SS_LIMIT}

    DURATION_RANGE_DICT = {'PnY': DURATION_PNY_LIMIT,
                           'PnM': DURATION_PNM_LIMIT,
                           'PnW': DURATION_PNW_LIMIT,
                           'PnD': DURATION_PND_LIMIT,
                           'TnH': DURATION_TNH_LIMIT,
                           'TnM': DURATION_TNM_LIMIT,
                           'TnS': DURATION_TNS_LIMIT}

    @classmethod
    def build_date(cls, YYYY=None, MM=None, DD=None, Www=None, D=None,
                   DDD=None):
        YYYY, MM, DD, Www, D, DDD = cls.range_check_date(YYYY, MM, DD, Www, D, DDD)

        if MM is None:
            MM = 1

        if DD is None:
            DD = 1

        if DDD is not None:
            return PythonTimeBuilder._build_ordinal_date(YYYY, DDD)

        if Www is not None:
            return PythonTimeBuilder._build_week_date(YYYY, Www,
                                                      isoday=D)

        return datetime.date(YYYY, MM, DD)

    @classmethod
    def build_time(cls, hh=None, mm=None, ss=None, tz=None):
        #Builds a time from the given parts, handling fractional arguments
        #where necessary
        hours = 0
        minutes = 0
        seconds = 0
        microseconds = 0

        hh, mm, ss, tz = cls.range_check_time(hh, mm, ss, tz)

        if type(hh) is FractionalComponent:
            hours = hh.principal
            microseconds = hh.microsecondremainder
        elif hh is not None:
            hours = hh

        if type(mm) is FractionalComponent:
            minutes = mm.principal
            microseconds = mm.microsecondremainder
        elif mm is not None:
            minutes = mm

        if type(ss) is FractionalComponent:
            seconds = ss.principal
            microseconds = ss.microsecondremainder
        elif ss is not None:
            seconds = ss

        hours, minutes, seconds, microseconds = PythonTimeBuilder._distribute_microseconds(microseconds, (hours, minutes, seconds), (MICROSECONDS_PER_HOUR, MICROSECONDS_PER_MINUTE, MICROSECONDS_PER_SECOND))

        #Move midnight into range
        if hours == 24:
            hours = 0

        #Datetimes don't handle fractional components, so we use a timedelta
        if tz is not None:
            return (datetime.datetime(1, 1, 1,
                                      hour=hours,
                                      minute=minutes,
                                      tzinfo=cls._build_object(tz))
                    + datetime.timedelta(seconds=seconds,
                                         microseconds=microseconds)
                   ).timetz()

        return (datetime.datetime(1, 1, 1,
                                  hour=hours,
                                  minute=minutes)
                + datetime.timedelta(seconds=seconds,
                                     microseconds=microseconds)
               ).time()

    @classmethod
    def build_datetime(cls, date, time):
        return datetime.datetime.combine(cls._build_object(date),
                                         cls._build_object(time))

    @classmethod
    def build_duration(cls, PnY=None, PnM=None, PnW=None, PnD=None, TnH=None,
                       TnM=None, TnS=None):
        #PnY and PnM will be distributed to PnD, microsecond remainder to TnS
        PnY, PnM, PnW, PnD, TnH, TnM, TnS = cls.range_check_duration(PnY, PnM, PnW, PnD, TnH, TnM, TnS)

        seconds = TnS.principal
        microseconds = TnS.microsecondremainder

        return datetime.timedelta(days=PnD,
                                  seconds=seconds,
                                  microseconds=microseconds,
                                  minutes=TnM,
                                  hours=TnH,
                                  weeks=PnW)

    @classmethod
    def build_interval(cls, start=None, end=None, duration=None):
        start, end, duration = cls.range_check_interval(start, end, duration)

        if start is not None and end is not None:
            #<start>/<end>
            startobject = cls._build_object(start)
            endobject = cls._build_object(end)

            return (startobject, endobject)

        durationobject = cls._build_object(duration)

        #Determine if datetime promotion is required
        datetimerequired = (duration.TnH is not None
                            or duration.TnM is not None
                            or duration.TnS is not None
                            or durationobject.seconds != 0
                            or durationobject.microseconds != 0)

        if end is not None:
            #<duration>/<end>
            endobject = cls._build_object(end)

            #Range check
            if type(end) is DateTuple and datetimerequired is True:
                #<end> is a date, and <duration> requires datetime resolution
                return (endobject,
                        cls.build_datetime(end, TupleBuilder.build_time())
                        - durationobject)

            return (endobject,
                    endobject
                    - durationobject)

        #<start>/<duration>
        startobject = cls._build_object(start)

        #Range check
        if type(start) is DateTuple and datetimerequired is True:
            #<start> is a date, and <duration> requires datetime resolution
            return (startobject,
                    cls.build_datetime(start, TupleBuilder.build_time())
                    + durationobject)

        return (startobject,
                startobject
                + durationobject)

    @classmethod
    def build_repeating_interval(cls, R=None, Rnn=None, interval=None):
        startobject = None
        endobject = None

        R, Rnn, interval = cls.range_check_repeating_interval(R, Rnn, interval)

        if interval.start is not None:
            startobject = cls._build_object(interval.start)

        if interval.end is not None:
            endobject = cls._build_object(interval.end)

        if interval.duration is not None:
            durationobject = cls._build_object(interval.duration)
        else:
            durationobject = endobject - startobject

        if R is True:
            if startobject is not None:
                return cls._date_generator_unbounded(startobject,
                                                     durationobject)

            return cls._date_generator_unbounded(endobject,
                                                 -durationobject)

        iterations = int(Rnn)

        if startobject is not None:
            return cls._date_generator(startobject, durationobject, iterations)

        return cls._date_generator(endobject, -durationobject, iterations)

    @classmethod
    def build_timezone(cls, negative=None, Z=None, hh=None, mm=None, name=''):
        negative, Z, hh, mm, name = cls.range_check_timezone(negative, Z, hh, mm, name)

        if Z is True:
            #Z -> UTC
            return UTCOffset(name='UTC', minutes=0)

        tzhour = int(hh)

        if mm is not None:
            tzminute = int(mm)
        else:
            tzminute = 0

        if negative is True:
            return UTCOffset(name=name, minutes=-(tzhour * 60 + tzminute))

        return UTCOffset(name=name, minutes=tzhour * 60 + tzminute)

    @classmethod
    def range_check_duration(cls, PnY=None, PnM=None, PnW=None, PnD=None,
                             TnH=None, TnM=None, TnS=None, rangedict=None):
        years = 0
        months = 0
        days = 0
        weeks = 0
        hours = 0
        minutes = 0
        seconds = 0
        microseconds = 0

        PnY, PnM, PnW, PnD, TnH, TnM, TnS = BaseTimeBuilder.range_check_duration(PnY, PnM, PnW, PnD, TnH, TnM, TnS, rangedict=cls.DURATION_RANGE_DICT)

        if PnY is not None:
            if type(PnY) is FractionalComponent:
                years = PnY.principal
                microseconds = PnY.microsecondremainder
            else:
                years = PnY

            if years * DAYS_PER_YEAR > TIMEDELTA_MAX_DAYS:
                raise YearOutOfBoundsError('Duration exceeds maximum timedelta size.')

        if PnM is not None:
            if type(PnM) is FractionalComponent:
                months = PnM.principal
                microseconds = PnM.microsecondremainder
            else:
                months = PnM

            if months * DAYS_PER_MONTH > TIMEDELTA_MAX_DAYS:
                raise MonthOutOfBoundsError('Duration exceeds maximum timedelta size.')

        if PnW is not None:
            if type(PnW) is FractionalComponent:
                weeks = PnW.principal
                microseconds = PnW.microsecondremainder
            else:
                weeks = PnW

            if weeks * DAYS_PER_WEEK > TIMEDELTA_MAX_DAYS:
                raise WeekOutOfBoundsError('Duration exceeds maximum timedelta size.')

        if PnD is not None:
            if type(PnD) is FractionalComponent:
                days = PnD.principal
                microseconds = PnD.microsecondremainder
            else:
                days = PnD

            if days > TIMEDELTA_MAX_DAYS:
                raise DayOutOfBoundsError('Duration exceeds maximum timedelta size.')

        if TnH is not None:
            if type(TnH) is FractionalComponent:
                hours = TnH.principal
                microseconds = TnH.microsecondremainder
            else:
                hours = TnH

            if hours // HOURS_PER_DAY > TIMEDELTA_MAX_DAYS:
                raise HoursOutOfBoundsError('Duration exceeds maximum timedelta size.')

        if TnM is not None:
            if type(TnM) is FractionalComponent:
                minutes = TnM.principal
                microseconds = TnM.microsecondremainder
            else:
                minutes = TnM

            if minutes // MINUTES_PER_DAY > TIMEDELTA_MAX_DAYS:
                raise MinutesOutOfBoundsError('Duration exceeds maximum timedelta size.')

        if TnS is not None:
            if type(TnS) is FractionalComponent:
                seconds = TnS.principal
                microseconds = TnS.microsecondremainder
            else:
                seconds = TnS

            if seconds // SECONDS_PER_DAY > TIMEDELTA_MAX_DAYS:
                raise SecondsOutOfBoundsError('Duration exceeds maximum timedelta size.')

        years, months, weeks, days, hours, minutes, seconds, microseconds = PythonTimeBuilder._distribute_microseconds(microseconds, (years, months, weeks, days, hours, minutes, seconds), (MICROSECONDS_PER_YEAR, MICROSECONDS_PER_MONTH, MICROSECONDS_PER_WEEK, MICROSECONDS_PER_DAY, MICROSECONDS_PER_HOUR, MICROSECONDS_PER_MINUTE, MICROSECONDS_PER_SECOND))

        #Note that weeks can be handled without conversion to days
        totaldays = years * DAYS_PER_YEAR + months * DAYS_PER_MONTH + days

        #Check against timedelta limits
        if totaldays + weeks * DAYS_PER_WEEK + hours // HOURS_PER_DAY + minutes // MINUTES_PER_DAY + seconds // SECONDS_PER_DAY > TIMEDELTA_MAX_DAYS:
            raise DayOutOfBoundsError('Duration exceeds maximum timedelta size.')

        return (None, None, weeks, totaldays, hours, minutes, FractionalComponent(seconds, microseconds))

    @classmethod
    def range_check_interval(cls, start=None, end=None, duration=None):
        #Handles concise format, range checks any potential durations
        if start is not None and end is not None:
            #<start>/<end>
            #Handle concise format
            if cls._is_interval_end_concise(end) is True:
                end = cls._combine_concise_interval_tuples(start, end)

            return (start, end, duration)

        durationobject = cls._build_object(duration)

        if end is not None:
            #<duration>/<end>
            endobject = cls._build_object(end)

            #Range check
            if type(end) is DateTuple:
                enddatetime = cls.build_datetime(end, TupleBuilder.build_time())

                if enddatetime - datetime.datetime.min < durationobject:
                    raise YearOutOfBoundsError('Interval end less than minimium date.')
            else:
                mindatetime = datetime.datetime.min

                if end.time.tz is not None:
                    mindatetime = mindatetime.replace(tzinfo=endobject.tzinfo)

                if endobject - mindatetime < durationobject:
                    raise YearOutOfBoundsError('Interval end less than minimium date.')
        else:
            #<start>/<duration>
            startobject = cls._build_object(start)

            #Range check
            if type(start) is DateTuple:
                startdatetime = cls.build_datetime(start, TupleBuilder.build_time())

                if datetime.datetime.max - startdatetime < durationobject:
                    raise YearOutOfBoundsError('Interval end greater than maximum date.')
            else:
                maxdatetime = datetime.datetime.max

                if start.time.tz is not None:
                    maxdatetime = maxdatetime.replace(tzinfo=startobject.tzinfo)

                if maxdatetime - startobject < durationobject:
                    raise YearOutOfBoundsError('Interval end greater than maximum date.')

        return (start, end, duration)

    @staticmethod
    def _build_week_date(isoyear, isoweek, isoday=None):
        if isoday is None:
            return (PythonTimeBuilder._iso_year_start(isoyear)
                    + datetime.timedelta(weeks=isoweek - 1))

        return (PythonTimeBuilder._iso_year_start(isoyear)
                + datetime.timedelta(weeks=isoweek - 1, days=isoday - 1))

    @staticmethod
    def _build_ordinal_date(isoyear, isoday):
        #Day of year to a date
        #https://stackoverflow.com/questions/2427555/python-question-year-and-day-of-year-to-date
        builtdate = (datetime.date(isoyear, 1, 1)
                     + datetime.timedelta(days=isoday - 1))

        return builtdate

    @staticmethod
    def _iso_year_start(isoyear):
        #Given an ISO year, returns the equivalent of the start of the year
        #on the Gregorian calendar (which is used by Python)
        #Stolen from:
        #http://stackoverflow.com/questions/304256/whats-the-best-way-to-find-the-inverse-of-datetime-isocalendar

        #Determine the location of the 4th of January, the first week of
        #the ISO year is the week containing the 4th of January
        #http://en.wikipedia.org/wiki/ISO_week_date
        fourth_jan = datetime.date(isoyear, 1, 4)

        #Note the conversion from ISO day (1 - 7) and Python day (0 - 6)
        delta = datetime.timedelta(days=fourth_jan.isoweekday() - 1)

        #Return the start of the year
        return fourth_jan - delta

    @staticmethod
    def _date_generator(startdate, timedelta, iterations):
        currentdate = startdate
        currentiteration = 0

        while currentiteration < iterations:
            yield currentdate

            #Update the values
            currentdate += timedelta
            currentiteration += 1

    @staticmethod
    def _date_generator_unbounded(startdate, timedelta):
        currentdate = startdate

        while True:
            yield currentdate

            #Update the value
            currentdate += timedelta

    @staticmethod
    def _distribute_microseconds(todistribute, recipients, reductions):
        #Given a number of microseconds as int, a tuple of ints length n
        #to distribute to, and a tuple of ints length n to divide todistribute
        #by (from largest to smallest), returns a tuple of length n + 1, with
        #todistribute divided across recipients using the reductions, with
        #the final remainder returned as the final tuple member
        results = []

        remainder = todistribute

        for index, reduction in enumerate(reductions):
            additional, remainder = divmod(remainder, reduction)

            results.append(recipients[index] + additional)

        #Always return the remaining microseconds
        results.append(remainder)

        return tuple(results)



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

#Import the main parsing functions so they are readily available
from aniso8601.time import (parse_datetime, parse_time,
                            get_datetime_resolution, get_time_resolution)
from aniso8601.date import parse_date, get_date_resolution
from aniso8601.duration import parse_duration, get_duration_resolution
from aniso8601.interval import (parse_interval, parse_repeating_interval,
                                get_interval_resolution, get_repeating_interval_resolution)

__version__ = "9.0.0"



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import calendar

from aniso8601.exceptions import (DayOutOfBoundsError, MidnightBoundsError,
                                  MinutesOutOfBoundsError, MonthOutOfBoundsError,
                                  HoursOutOfBoundsError, ISOFormatError,
                                  LeapSecondError, SecondsOutOfBoundsError,
                                  WeekOutOfBoundsError, YearOutOfBoundsError)
from collections import namedtuple

DateTuple = namedtuple('Date', ['YYYY', 'MM', 'DD', 'Www', 'D', 'DDD'])
TimeTuple = namedtuple('Time', ['hh', 'mm', 'ss', 'tz'])
DatetimeTuple = namedtuple('Datetime', ['date', 'time'])
DurationTuple = namedtuple('Duration', ['PnY', 'PnM', 'PnW', 'PnD',
                                        'TnH', 'TnM', 'TnS'])
IntervalTuple = namedtuple('Interval', ['start', 'end', 'duration'])
RepeatingIntervalTuple = namedtuple('RepeatingInterval', ['R', 'Rnn',
                                                          'interval'])
TimezoneTuple = namedtuple('Timezone', ['negative', 'Z', 'hh', 'mm', 'name'])

Limit = namedtuple('Limit', ['casterrorstring', 'min', 'max',
                             'rangeexception', 'rangeerrorstring',
                             'rangefunc'])

def cast(value, castfunction, caughtexceptions=(ValueError,),
         thrownexception=ISOFormatError, thrownmessage=None):
    try:
        result = castfunction(value)
    except caughtexceptions:
        raise thrownexception(thrownmessage)

    return result

def range_check(valuestr, limit):
    #Returns cast value if in range, raises defined exceptions on failure
    if valuestr is None:
        return None

    if '.' in valuestr:
        castfunc = float
    else:
        castfunc = int

    value = cast(valuestr, castfunc, thrownmessage=limit.casterrorstring)

    if limit.min is not None and value < limit.min:
        raise limit.rangeexception(limit.rangeerrorstring)

    if limit.max is not None and value > limit.max:
        raise limit.rangeexception(limit.rangeerrorstring)

    return value

class BaseTimeBuilder(object):
    #Limit tuple format cast function, cast error string,
    #lower limit, upper limit, limit error string
    DATE_YYYY_LIMIT = Limit('Invalid year string.',
                            0000, 9999, YearOutOfBoundsError,
                            'Year must be between 1..9999.',
                            range_check)
    DATE_MM_LIMIT = Limit('Invalid month string.',
                          1, 12, MonthOutOfBoundsError,
                          'Month must be between 1..12.',
                          range_check)
    DATE_DD_LIMIT = Limit('Invalid day string.',
                          1, 31, DayOutOfBoundsError,
                          'Day must be between 1..31.',
                          range_check)
    DATE_WWW_LIMIT = Limit('Invalid week string.',
                           1, 53, WeekOutOfBoundsError,
                           'Week number must be between 1..53.',
                           range_check)
    DATE_D_LIMIT = Limit('Invalid weekday string.',
                         1, 7, DayOutOfBoundsError,
                         'Weekday number must be between 1..7.',
                         range_check)
    DATE_DDD_LIMIT = Limit('Invalid ordinal day string.',
                           1, 366, DayOutOfBoundsError,
                           'Ordinal day must be between 1..366.',
                           range_check)
    TIME_HH_LIMIT = Limit('Invalid hour string.',
                          0, 24, HoursOutOfBoundsError,
                          'Hour must be between 0..24 with '
                          '24 representing midnight.',
                          range_check)
    TIME_MM_LIMIT = Limit('Invalid minute string.',
                          0, 59, MinutesOutOfBoundsError,
                          'Minute must be between 0..59.',
                          range_check)
    TIME_SS_LIMIT = Limit('Invalid second string.',
                          0, 60, SecondsOutOfBoundsError,
                          'Second must be between 0..60 with '
                          '60 representing a leap second.',
                          range_check)
    TZ_HH_LIMIT = Limit('Invalid timezone hour string.',
                        0, 23, HoursOutOfBoundsError,
                        'Hour must be between 0..23.',
                        range_check)
    TZ_MM_LIMIT = Limit('Invalid timezone minute string.',
                        0, 59, MinutesOutOfBoundsError,
                        'Minute must be between 0..59.',
                        range_check)
    DURATION_PNY_LIMIT = Limit('Invalid year duration string.',
                               0, None, ISOFormatError,
                               'Duration years component must be positive.',
                               range_check)
    DURATION_PNM_LIMIT = Limit('Invalid month duration string.',
                               0, None, ISOFormatError,
                               'Duration months component must be positive.',
                               range_check)
    DURATION_PNW_LIMIT = Limit('Invalid week duration string.',
                               0, None, ISOFormatError,
                               'Duration weeks component must be positive.',
                               range_check)
    DURATION_PND_LIMIT = Limit('Invalid day duration string.',
                               0, None, ISOFormatError,
                               'Duration days component must be positive.',
                               range_check)
    DURATION_TNH_LIMIT = Limit('Invalid hour duration string.',
                               0, None, ISOFormatError,
                               'Duration hours component must be positive.',
                               range_check)
    DURATION_TNM_LIMIT = Limit('Invalid minute duration string.',
                               0, None, ISOFormatError,
                               'Duration minutes component must be positive.',
                               range_check)
    DURATION_TNS_LIMIT = Limit('Invalid second duration string.',
                               0, None, ISOFormatError,
                               'Duration seconds component must be positive.',
                               range_check)
    INTERVAL_RNN_LIMIT = Limit('Invalid duration repetition string.',
                               0, None, ISOFormatError,
                               'Duration repetition count must be positive.',
                               range_check)

    DATE_RANGE_DICT = {'YYYY': DATE_YYYY_LIMIT, 'MM': DATE_MM_LIMIT,
                       'DD': DATE_DD_LIMIT, 'Www': DATE_WWW_LIMIT,
                       'D': DATE_D_LIMIT, 'DDD': DATE_DDD_LIMIT}

    TIME_RANGE_DICT = {'hh': TIME_HH_LIMIT, 'mm': TIME_MM_LIMIT, 'ss': TIME_SS_LIMIT}

    DURATION_RANGE_DICT = {'PnY': DURATION_PNY_LIMIT,
                           'PnM': DURATION_PNM_LIMIT,
                           'PnW': DURATION_PNW_LIMIT,
                           'PnD': DURATION_PND_LIMIT,
                           'TnH': DURATION_TNH_LIMIT,
                           'TnM': DURATION_TNM_LIMIT,
                           'TnS': DURATION_TNS_LIMIT}

    REPEATING_INTERVAL_RANGE_DICT = {'Rnn': INTERVAL_RNN_LIMIT}

    TIMEZONE_RANGE_DICT = {'hh': TZ_HH_LIMIT, 'mm': TZ_MM_LIMIT}

    LEAP_SECONDS_SUPPORTED = False

    @classmethod
    def build_date(cls, YYYY=None, MM=None, DD=None, Www=None, D=None,
                   DDD=None):
        raise NotImplementedError

    @classmethod
    def build_time(cls, hh=None, mm=None, ss=None, tz=None):
        raise NotImplementedError

    @classmethod
    def build_datetime(cls, date, time):
        raise NotImplementedError

    @classmethod
    def build_duration(cls, PnY=None, PnM=None, PnW=None, PnD=None, TnH=None,
                       TnM=None, TnS=None):
        raise NotImplementedError

    @classmethod
    def build_interval(cls, start=None, end=None, duration=None):
        #start, end, and duration are all tuples
        raise NotImplementedError

    @classmethod
    def build_repeating_interval(cls, R=None, Rnn=None, interval=None):
        #interval is a tuple
        raise NotImplementedError

    @classmethod
    def build_timezone(cls, negative=None, Z=None, hh=None, mm=None, name=''):
        raise NotImplementedError

    @classmethod
    def range_check_date(cls, YYYY=None, MM=None, DD=None,
                         Www=None, D=None, DDD=None, rangedict=None):
        if rangedict is None:
            rangedict = cls.DATE_RANGE_DICT

        if 'YYYY' in rangedict:
            YYYY = rangedict['YYYY'].rangefunc(YYYY, rangedict['YYYY'])

        if 'MM' in rangedict:
            MM = rangedict['MM'].rangefunc(MM, rangedict['MM'])

        if 'DD' in rangedict:
            DD = rangedict['DD'].rangefunc(DD, rangedict['DD'])

        if 'Www' in rangedict:
            Www = rangedict['Www'].rangefunc(Www, rangedict['Www'])

        if 'D' in rangedict:
            D = rangedict['D'].rangefunc(D, rangedict['D'])

        if 'DDD' in rangedict:
            DDD = rangedict['DDD'].rangefunc(DDD, rangedict['DDD'])

        if DD is not None:
            #Check calendar
            if DD > calendar.monthrange(YYYY, MM)[1]:
                raise DayOutOfBoundsError('{0} is out of range for {1}-{2}'.format(DD, YYYY, MM))

        if DDD is not None:
            if calendar.isleap(YYYY) is False and DDD == 366:
                raise DayOutOfBoundsError('{0} is only valid for leap year.'.format(DDD))

        return (YYYY, MM, DD, Www, D, DDD)

    @classmethod
    def range_check_time(cls, hh=None, mm=None, ss=None, tz=None, rangedict=None):
        #Used for midnight and leap second handling
        midnight = False #Handle hh = '24' specially

        if rangedict is None:
            rangedict = cls.TIME_RANGE_DICT

        if 'hh' in rangedict:
            try:
                hh = rangedict['hh'].rangefunc(hh, rangedict['hh'])
            except HoursOutOfBoundsError as e:
                if float(hh) > 24 and float(hh) < 25:
                    raise MidnightBoundsError('Hour 24 may only represent midnight.')

                raise e

        if 'mm' in rangedict:
            mm = rangedict['mm'].rangefunc(mm, rangedict['mm'])

        if 'ss' in rangedict:
            ss = rangedict['ss'].rangefunc(ss, rangedict['ss'])

        if hh is not None and hh == 24:
            midnight = True

        #Handle midnight range
        if midnight is True and ((mm is not None and mm != 0) or (ss is not None and ss != 0)):
            raise MidnightBoundsError('Hour 24 may only represent midnight.')

        if cls.LEAP_SECONDS_SUPPORTED is True:
            if hh != 23 and mm != 59 and ss == 60:
                raise cls.TIME_SS_LIMIT.rangeexception(cls.TIME_SS_LIMIT.rangeerrorstring)
        else:
            if hh == 23 and mm == 59 and ss == 60:
                #https://bitbucket.org/nielsenb/aniso8601/issues/10/sub-microsecond-precision-in-durations-is
                raise LeapSecondError('Leap seconds are not supported.')

            if ss == 60:
                raise cls.TIME_SS_LIMIT.rangeexception(cls.TIME_SS_LIMIT.rangeerrorstring)

        return (hh, mm, ss, tz)

    @classmethod
    def range_check_duration(cls, PnY=None, PnM=None, PnW=None, PnD=None,
                             TnH=None, TnM=None, TnS=None, rangedict=None):
        if rangedict is None:
            rangedict = cls.DURATION_RANGE_DICT

        if 'PnY' in rangedict:
            PnY = rangedict['PnY'].rangefunc(PnY, rangedict['PnY'])

        if 'PnM' in rangedict:
            PnM = rangedict['PnM'].rangefunc(PnM, rangedict['PnM'])

        if 'PnW' in rangedict:
            PnW = rangedict['PnW'].rangefunc(PnW, rangedict['PnW'])

        if 'PnD' in rangedict:
            PnD = rangedict['PnD'].rangefunc(PnD, rangedict['PnD'])

        if 'TnH' in rangedict:
            TnH = rangedict['TnH'].rangefunc(TnH, rangedict['TnH'])

        if 'TnM' in rangedict:
            TnM = rangedict['TnM'].rangefunc(TnM, rangedict['TnM'])

        if 'TnS' in rangedict:
            TnS = rangedict['TnS'].rangefunc(TnS, rangedict['TnS'])

        return (PnY, PnM, PnW, PnD, TnH, TnM, TnS)

    @classmethod
    def range_check_repeating_interval(cls, R=None, Rnn=None, interval=None,
                                       rangedict=None):
        if rangedict is None:
            rangedict = cls.REPEATING_INTERVAL_RANGE_DICT

        if 'Rnn' in rangedict:
            Rnn = rangedict['Rnn'].rangefunc(Rnn, rangedict['Rnn'])

        return (R, Rnn, interval)

    @classmethod
    def range_check_timezone(cls, negative=None, Z=None, hh=None, mm=None,
                             name='', rangedict=None):
        if rangedict is None:
            rangedict = cls.TIMEZONE_RANGE_DICT

        if 'hh' in rangedict:
            hh = rangedict['hh'].rangefunc(hh, rangedict['hh'])

        if 'mm' in rangedict:
            mm = rangedict['mm'].rangefunc(mm, rangedict['mm'])

        return (negative, Z, hh, mm, name)

    @classmethod
    def _build_object(cls, parsetuple):
        #Given a TupleBuilder tuple, build the correct object
        if type(parsetuple) is DateTuple:
            return cls.build_date(YYYY=parsetuple.YYYY, MM=parsetuple.MM,
                                  DD=parsetuple.DD, Www=parsetuple.Www,
                                  D=parsetuple.D, DDD=parsetuple.DDD)

        if type(parsetuple) is TimeTuple:
            return cls.build_time(hh=parsetuple.hh, mm=parsetuple.mm,
                                  ss=parsetuple.ss, tz=parsetuple.tz)

        if type(parsetuple) is DatetimeTuple:
            return cls.build_datetime(parsetuple.date, parsetuple.time)

        if type(parsetuple) is DurationTuple:
            return cls.build_duration(PnY=parsetuple.PnY, PnM=parsetuple.PnM,
                                      PnW=parsetuple.PnW, PnD=parsetuple.PnD,
                                      TnH=parsetuple.TnH, TnM=parsetuple.TnM,
                                      TnS=parsetuple.TnS)

        if type(parsetuple) is IntervalTuple:
            return cls.build_interval(start=parsetuple.start, end=parsetuple.end,
                                      duration=parsetuple.duration)

        if type(parsetuple) is RepeatingIntervalTuple:
            return cls.build_repeating_interval(R=parsetuple.R,
                                                Rnn=parsetuple.Rnn,
                                                interval=parsetuple.interval)

        return cls.build_timezone(negative=parsetuple.negative, Z=parsetuple.Z,
                                  hh=parsetuple.hh, mm=parsetuple.mm,
                                  name=parsetuple.name)

    @classmethod
    def _is_interval_end_concise(cls, endtuple):
        if type(endtuple) is TimeTuple:
            return True

        if type(endtuple) is DatetimeTuple:
            enddatetuple = endtuple.date
        else:
            enddatetuple = endtuple

        if enddatetuple.YYYY is None:
            return True

        return False

    @classmethod
    def _combine_concise_interval_tuples(cls, starttuple, conciseendtuple):
        starttimetuple = None
        startdatetuple = None

        endtimetuple = None
        enddatetuple = None

        if type(starttuple) is DateTuple:
            startdatetuple = starttuple
        else:
            #Start is a datetime
            starttimetuple = starttuple.time
            startdatetuple = starttuple.date

        if type(conciseendtuple) is DateTuple:
            enddatetuple = conciseendtuple
        elif type(conciseendtuple) is DatetimeTuple:
            enddatetuple = conciseendtuple.date
            endtimetuple = conciseendtuple.time
        else:
            #Time
            endtimetuple = conciseendtuple

        if enddatetuple is not None:
            if enddatetuple.YYYY is None and enddatetuple.MM is None:
                newenddatetuple = DateTuple(YYYY=startdatetuple.YYYY, MM=startdatetuple.MM,
                                            DD=enddatetuple.DD, Www=enddatetuple.Www,
                                            D=enddatetuple.D, DDD=enddatetuple.DDD)
            else:
                newenddatetuple = DateTuple(YYYY=startdatetuple.YYYY, MM=enddatetuple.MM,
                                            DD=enddatetuple.DD, Www=enddatetuple.Www,
                                            D=enddatetuple.D, DDD=enddatetuple.DDD)

        if ((starttimetuple is not None and starttimetuple.tz is not None) and
            (endtimetuple is not None and endtimetuple.tz != starttimetuple.tz)):
            #Copy the timezone across
            endtimetuple = TimeTuple(hh=endtimetuple.hh, mm=endtimetuple.mm,
                                     ss=endtimetuple.ss, tz=starttimetuple.tz)

        if enddatetuple is not None and endtimetuple is None:
            return newenddatetuple

        if enddatetuple is not None and endtimetuple is not None:
            return TupleBuilder.build_datetime(newenddatetuple, endtimetuple)

        return TupleBuilder.build_datetime(startdatetuple, endtimetuple)

class TupleBuilder(BaseTimeBuilder):
    #Builder used to return the arguments as a tuple, cleans up some parse methods
    @classmethod
    def build_date(cls, YYYY=None, MM=None, DD=None, Www=None, D=None,
                   DDD=None):

        return DateTuple(YYYY, MM, DD, Www, D, DDD)

    @classmethod
    def build_time(cls, hh=None, mm=None, ss=None, tz=None):
        return TimeTuple(hh, mm, ss, tz)

    @classmethod
    def build_datetime(cls, date, time):
        return DatetimeTuple(date, time)

    @classmethod
    def build_duration(cls, PnY=None, PnM=None, PnW=None, PnD=None, TnH=None,
                       TnM=None, TnS=None):

        return DurationTuple(PnY, PnM, PnW, PnD, TnH, TnM, TnS)

    @classmethod
    def build_interval(cls, start=None, end=None, duration=None):
        return IntervalTuple(start, end, duration)

    @classmethod
    def build_repeating_interval(cls, R=None, Rnn=None, interval=None):
        return RepeatingIntervalTuple(R, Rnn, interval)

    @classmethod
    def build_timezone(cls, negative=None, Z=None, hh=None, mm=None, name=''):
        return TimezoneTuple(negative, Z, hh, mm, name)
